#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-
#pylint: disable-msg=E1101,C0103,R0902,R0903
#Author: Valentin Kuznetsov

"""
ReleaseManager is responsible for setting up CMSSW area and run cmsRun job 
to extract run/events from given LFN/PFN list
"""

import os
import sets
import stat
import time
import popen2
import tempfile
import threading
import ConfigParser

import DBSInteraction
from ConfiguredObject import ConfiguredObject
from ThreadPool import ThreadPool
from Utils import natsort, uniqueList, iparser, edmconfig, walktree, cleanup

class CMSRunManager(ConfiguredObject):
    """Manage cmsRun jobs"""

    def __init__(self, area, release, config):
        super(CMSRunManager, self).__init__()
        self.area = area         # CMSSW release area to be used, setup by scram
        self.release = release   # CMSSW release to be used
        self.config  = config    # cmsRun configuration
        self.process = None      # cmsRun process Popen4 object
        self.tmpfilename = None  # cmsRun configuration file name
        self.release_to_use = '' # CMSSW released used to produce output file
        self.outfilename    = '' # output file name
        self.inputparams = None  # input parameters: dataset, eventset

        # find list of available release on our system
        cmd = "scram list -c CMSSW | awk '{print $2}'"
        res = popen2.Popen4(cmd)
        res.wait()
        releases = res.fromchild.read()
        self.releases = natsort(list(releases.split()))
        
        # find best matched release wrt one we need to use
        for rel in self.releases:
            if  rel <= self.release:
                self.release_to_use = rel

    def setup_cmssw(self):
        """setup CMSSW area and run scram"""
        print "setting up cmssw"
        cdir = os.getcwd()
        os.chdir(self.area)
        cmd = 'scramv1 project CMSSW %s' % self.release_to_use
        os.system(cmd)
        os.chdir(cdir)

    def start(self):
        """start cmsRun job"""
        self.log.info("Starting the cmsRun job")
        print "starting the cmsRun job"
        cdir = os.getcwd()
        self.setup_cmssw()
        os.chdir( os.path.join( self.area, self.release_to_use ))
        if  self.release_to_use < 'CMSSW_1_6':
            self.log.error("No support for release %s" % self.release_to_use)
            return # no support
        elif self.release_to_use < 'CMSSW_2_1':
            tmpfilename = tempfile.mktemp() + '.cfg'
        else:
            tmpfilename = tempfile.mktemp() + '.py'
        self.tmpfilename = tmpfilename
        tfile = open(tmpfilename, 'w')
        tfile.write(self.config)
        tfile.close()
        cmd = 'eval `scramv1 run -sh` && cmsRun %s' % tmpfilename
        print cmd
        process = popen2.Popen4(cmd)
        self.process = process
        os.chdir(cdir)
        print "Thread dying!"
        
    def process_output(self):
        """Read content of running process"""
        if  self.process:
            self.process.wait()
            res = self.process.fromchild.read()
            return res
        return
        
    def cancel(self):
        """cancel cmsRun job"""
        # TODO: implement kill -9
        if  self.process:
            pid = self.process.pid
            os.system('kill -9 %s' % pid)
            self.process = None
            os.remove(self.tmpfilename)
        return

    def status(self):
        """check status of cmsRun job"""
        status = -1
        if  self.process:
            status = self.process.poll()
            if  not status:
                self.process = None
                os.remove(self.tmpfilename)
        return status

    def filecontent(self):
        """Read content of output file using emdFileUtil"""
        cwd  = os.getcwd()
        area = os.path.join(self.area, self.release_to_use)
        os.chdir(area)
        cmd  = 'eval `scramv1 run -sh` && edmFileUtil -e -f file:%s' \
                % self.outfilename
        res  = popen2.Popen4(cmd)
        res.wait()
        content = res.fromchild.read()
        os.chdir(cwd)
        return content
    
    def nevents(self):
        """Use ROOT to check how many events are written in output file"""
        cdir = os.getcwd()
        tmpfilename = tempfile.mktemp() + '.C'
        template = """
using namespace std;

void %s() {
    TFile *f = TFile::Open("%s");
    TTree* tree;
    if ( tree = (TTree*)f->Get("Events") ) {
        cout << tree->GetEntries() << endl;
    }
}
    """ % (tmpfilename.split('/')[-1].replace('.C',''), self.outfilename)
        tfile = open(tmpfilename, 'w')
        tfile.write(template)
        tfile.close()
        area = os.path.join(self.area, self.release_to_use)
        os.chdir(area)
        cmd = 'eval `scramv1 run -sh` && root -l -n -q -b %s 2> /dev/null ' \
            '| tail -1' % tmpfilename
        res = popen2.Popen4(cmd)
        res.wait()
        content = res.fromchild.read()
        if  content.find('Processing') != -1:
            content = 0
        os.remove(tmpfilename)
        os.chdir(cdir)
        return content

class CMSRunThreadWrapper(ConfiguredObject):
    """
       this class handles setup of CMSSW area and running cmsRun job
    """
    def __init__(self, area, cp=None):

        self.cp           = None
        self.section      = "release_manager"
        super(CMSRunThreadWrapper, self).__init__(area)
        self.dbs = DBSInteraction.DBS('cms_dbs_prod_global')
        self._request     = {}
        self.max_cmsrun   = 3
        self.request_lock = threading.Lock()
        self.pool         = None

    def configure(self, cp):
        """configure wrapper"""
        self.cp = cp
        self.max_cmsrun = int(self.getOption("max_cmsrun", "5"))
        self.pool = ThreadPool(threads=self.max_cmsrun)
        
    def submit(self, user, requestdict):
        """
        Submit a cmsRun job over a provided request dictionary.
        """
        eventlist = requestdict.get('eventlist', [])
        dataset = requestdict.get('dataset', 'NONE')
        if not dataset:
            raise Exception("Request dictionary did not include dataset.")
        
        # Calculate the LFNs needed for this request.
        res  = sets.Set()
        for run, _, lumi in eventlist:
            res.update(self.dbs.getFiles(dataset=dataset, run=run, lumi=lumi))
        result = list(res)
        if not result:
            raise Exception("No files found!")

        cmsfs_location = None
        if self.cp:
            cmsfs_location = self.getOption("cmsfs_location", None)
        
        # Calculate the releases needed to process request.
        rellist = []
        lfnlist = []
        for lfn, _ in result:
            lfnlist.append(lfn)
            release = self.dbs.getRelease(lfn)
            if  not rellist.count(release):
                rellist.append(release)
        rellist.sort()

        # check that release list has 1 value
        if  len(rellist) > 1:
            msg  = 'Ambiguous request.\n'
            msg += 'Found multiple CMSSW releases for input parameter set.'
            raise Exception(msg)
        elif len(rellist) == 0:
            raise Exception("No release found for request.")
        
        release  = rellist[0]
        cmsswdir = self.getOption("cmssw_dir", "/tmp/cms_fs_files")
        odir     = os.path.join(cmsswdir, 'root_files')
        opath    = os.path.join(odir, str(user))
        try:
            os.makedirs(opath)
        except:
            pass
        walktree(odir, cleanup) # clean-up top-level dir
        outfilename = os.path.join(opath, \
                                   '%s.root' % str(time.time()).replace('.',''))

        if cmsfs_location:
            lfnlist = ['file:' + cmsfs_location + '/lfns' + i for i in lfnlist]

        config = edmconfig(release, lfnlist, eventlist, outfilename)
        if  not config:
            msg  = 'We unable to generate EDM config file based on provided ' \
                'information.<br/>'
            raise Exception(msg)
        if  os.path.isfile(outfilename):
            os.remove(outfilename)
        requestId = self.request(cmsswdir, release, config)
        if requestId:
            self.updateparams(requestId, outfilename, requestdict)
        return requestId

    def request(self, area, release, config):
        """
        Run cmsRun job over provided config file.
        Returns a requestId which can be used later for status checks.
        """
        if  len(self._request.keys()) == self.max_cmsrun:
            return None # to prevent from having too many cmsrun jobs on a node
        requestid = str(time.time()).replace('.','')
        self.request_lock.acquire()
        try:
            try:
                cmsrun = CMSRunManager(area, release, config)
                self._request[requestid] = cmsrun
                self.pool.queue(cmsrun)
            except RuntimeError, e:
                # unable to start new thread
                self.log.exception(e)
                requestid = None
        finally:
            self.request_lock.release()
        return requestid

    def status(self, requestid):
        """
           return status of cmsRun job
        """
        status = -1
        if  not requestid:
            return 'busy'
        self.request_lock.acquire()
        try:
            try:
                cmsrun = self._request[requestid]
                status = cmsrun.status()
                return status
            except (SystemExit, KeyboardInterrupt):
                raise
            except Exception, e:
                self.log.exception(e)
                return status
        finally:
            self.request_lock.release()
        return status
        
    def updateparams(self, requestid, outfilename, params):
        """update object with given input parameters for given requestid"""
        self._request[requestid].outfilename = outfilename
        self._request[requestid].inputparams = params
        # look-up for old files and clean them up
        # delete files older then 3 hours
        odir = os.path.split(outfilename)[0]
        for ifile in os.listdir(odir):
            ofile = os.path.join(odir, ifile)
            filestat = os.stat(ofile)
            if  filestat[stat.ST_MTIME] > (60*60*24):
                os.remove(ofile)
        
    def inputparams(self, requestid):
        """return input parameters for given requestid"""
        return self._request[requestid].inputparams
        
    def getoutfile(self, requestid):
        """return output file name for given requestid"""
        return self._request[requestid].outfilename
        
    def filecontent(self, requestid):
        """return file content of requestid"""
        return self._request[requestid].filecontent()
        
    def nevents(self, requestid):
        """return number of events in output file corresponding to requestid"""
        return self._request[requestid].nevents()

    def release_to_use(self, requestid):
        """return file content of requestid"""
        return self._request[requestid].release_to_use
        
    def config(self, requestid):
        """return file content of requestid"""
        return self._request[requestid].config

    def process_output(self, requestid):
        """return output of cmsRun for requestid"""
        return self._request[requestid].process_output()
        
    def clean(self, requestid):
        """clean internal cache about requestid"""
        del self._request[requestid]
        
    def cancel(self, requestid):
        """cancel requested cmsRun job"""
        self.request_lock.acquire()
        try:
            cmsrun = self._request[requestid]
            self.log.info("Sending a cancel request to cmsrun request=%s, " \
                          "filename=%s" % (requestid, cmsrun.outfilename))
            cmsrun.cancel()
            self.clean(requestid)
        finally:
            self.request_lock.release()

    def graceful_exit(self):
        """implement exit from thread pool"""
        self.pool.drain()
        self.pool.join()
#
# main
#
def test(dbsmgr, mgr, idir, dataset, eventset):
    """test request"""
    eventlist = [x for x in iparser(eventset)]
    res = []
    for run, evt, lumi in eventlist:
        res = dbsmgr.getFiles(run, dataset, evt, lumi)
    result = uniqueList(res)
    if  result:
        lfnlist = []
        for lfn, _ in result:
            lfnlist.append(lfn)
            release = dbsmgr.getRelease(lfn)
        outfilename = idir + '/root_files/valya/test.root'
        config = edmconfig(release, lfnlist, eventlist, outfilename)
        requestid = mgr.request(dir, release, config)
        if  not requestid:
            print "#### UNABLE to start thread for", dataset, eventset
        return requestid
    return None, None

def check_status(mgr, requestid):
    """test status"""
    counter = 0
    outfilename = mgr.getoutfile(requestid)
    while 1:
        status = mgr.status(requestid)
        if  status == 'busy':
            break
        if  os.path.isfile(outfilename):
            print '%s %s' % (os.stat(outfilename)[stat.ST_SIZE], outfilename)
        else:
            print 'cmsRun job is launched, please wait, job id: %s, status %s'\
                 % (outfilename, status)
        if  not status:
            print "Finish!!!", outfilename
            break
        counter += 1
        time.sleep(3)
        if counter > 25:
            break

def main():
    """main test"""
    idir = '/data/vk/filemover/cmssw'
    mgr = CMSRunThreadWrapper(idir)
    cp = ConfigParser.ConfigParser()
    cp.add_section("release_manager")
    cp.set("release_manager", "max_cmsrun", "2")
    mgr.configure(cp)
    import thread
    import DBSInteraction
    dbsmgr = DBSInteraction.DBS('cms_dbs_prod_global')
    # test CMSSW_1_6_X release series
    dataset = '/testbeam_HCalEcalCombined/h2tb2007_default_v1/DIGI-RECO'
    eventset = "16278 44000 10"
    requestid = test(dbsmgr, mgr, dir, dataset, eventset)
    print "####", requestid, mgr.getoutfile(requestid)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))
    # test CMSSW_2_1_X release series
    dataset  = '/Cosmics/Commissioning08-CRUZET4_v1/RECO'
    eventset = "58620 151578 6" 
    requestid = test(dbsmgr, mgr, dir, dataset, eventset)
    print "####", requestid, mgr.getoutfile(requestid)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))
    # use the same dataset but real event, this time we should fail since 
    # we only allow 2 cmsRun jobs
    eventset = "58620 151578 10" 
    requestid = test(dbsmgr, mgr, dir, dataset, eventset)
    print "####", requestid, mgr.getoutfile(requestid)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))
    time.sleep(90)

if __name__ == "__main__":
    main()
