
"""
Implementation of the CmsFs

The CmsFs is a FUSE-based file system which allows files to be read from remote
grid sites as if they were on the local file system.
"""

import os
import stat
import array
import errno
import logging
import threading
import ConfigParser

from fuse import Fuse #pylint:disable-msg=F0401
import fuse
fuse.fuse_python_api = (0, 2)

from ThreadPool import ThreadPool
from ConfiguredObject import ConfiguredObject
from FileManager import Server
from DBSInteraction import DBS
from CmsFsTransfers import BlockRequest, PartialFileFactory, BlockStore
from CmsFsGridftp import GridftpPartialRunner
from FileLookup import LookupServer
from Status import StatusCode

log = logging.getLogger("CmsFs.Cache")
fuse_log = logging.getLogger("CmsFs.FUSE")

class MyStat(fuse.Stat):
    
    """
    Stubs for a FUSE stat object.
    """
    
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0 
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0

class CmsFsFUSE(Fuse, ConfiguredObject):

    """
    The FUSE filesystem implementation for CmsFs.
    """

    def __init__(self, cp):
        self.cp = cp
        self.section = "gridftp_fuse"
        Fuse.__init__(self)
        ConfiguredObject.__init__(self)
        self._lock = threading.Lock()
        self._dbs_lock = threading.Lock()
        self._dbs = DBS()
        self._pf_factory = PartialFileFactory()
        self._block_store = BlockStore(self.cp, self._pf_factory)
        self.transfer_pool = None
        self.transfer_pool_size = -1
        self.partial_pool = None
        self.partial_pool_size = -1
        if not Server.is_configured():
            if not cp.has_section("file_manager"):
                cp.add_section("file_manager")
                cp.set("file_manager", "base_directory", "/tmp/cms_fs_files")
            Server.configure(cp)

    def configure(self, cp):
        """
        Configure the CmsFs FUSE filesystem based on a ConfigParser object.
        """
        self.cp = cp
        self.transfer_pool_size = int(self.getOption("transfer_pool_size", "5"))
        self.partial_pool_size = int(self.getOption("partial_pool_size", "5"))
        self.transfer_pool = ThreadPool(threads=self.transfer_pool_size)
        self.partial_pool = ThreadPool(threads=self.partial_pool_size)
        self._pf_factory.configure(cp)

    def getattr(self, path):
        """
        - st_mode (protection bits)
        - st_ino (inode number)
        - st_dev (device)
        - st_nlink (number of hard links)
        - st_uid (user ID of owner)
        - st_gid (group ID of owner)
        - st_size (size of file, in bytes)
        - st_atime (time of most recent access)
        - st_mtime (time of most recent content modification)
        - st_ctime (platform dependent; time of most recent metadata change 
                    on Unix, or the time of creation on Windows).
        """

        print '*** getattr', path
        st = MyStat()
        if path.startswith('/lfns/') and path.endswith('.root'):
            st.st_mode = stat.S_IFREG | 0444
            st.st_nlink = 1
            st.st_size = self.get_dbs_file(path[5:], lfn=True).size
        elif path in ['/', '/guids', '/runs', '/sites', '/lfns'] or \
                path.startswith('/lfns/'):
            st.st_mode = stat.S_IFDIR | 0755
            st.st_nlink = 2
        elif path.startswith('/guids/'):
            st.st_mode = stat.S_IFREG | 0444
            st.st_nlink = 1
            st.st_size = self.get_dbs_file(path[7:]).size
        else:
            return -errno.ENOENT
        #depth = getDepth(path) # depth of path, zero-based from root
        #pathparts = getParts(path) # the actual parts of the path

        return st

    def getdir(self, path):
        """
        return: [('file1', 0), ('file2', 0), ... ]
        """

        print '*** getdir', path
        return -errno.ENOSYS

    def readdir(self, path, offset):
        for r in '.', '..', 'runs', 'guids', 'sites', 'lfns':
            yield fuse.Direntry(r)

    def mythread ( self ):
        print '*** mythread'
        return -errno.ENOSYS

    def chmod ( self, path, mode ):
        """
        chmod; not functional
        """
        fuse_log.info('chmod %s %s' % (path, oct(mode)))
        return -errno.ENOSYS

    def chown ( self, path, uid, gid ):
        """
        chown; not functional
        """
        print '*** chown', path, uid, gid
        return -errno.ENOSYS

    def fsync ( self, path, isFsyncFile ):
        print '*** fsync', path, isFsyncFile
        return -errno.ENOSYS

    def link ( self, targetPath, linkPath ):
        print '*** link', targetPath, linkPath
        return -errno.ENOSYS

    def mkdir ( self, path, mode ):
        print '*** mkdir', path, oct(mode)
        return -errno.ENOSYS

    def mknod ( self, path, mode, dev ):
        """
        mknod; not functional
        """
        fuse_log.info('mknod %s %s %s' % (path, oct(mode), dev))
        return -errno.ENOSYS

    def open(self, path, flags):
        accmode = os.O_RDONLY | os.O_WRONLY | os.O_RDWR
        if (flags & accmode) != os.O_RDONLY:
            return -errno.EACCES
        parts = path.split('/')
        assert len(parts[0]) == 0
        if len(parts) == 3 and parts[1] == 'guids':
            name = self.get_dbs_file(parts[2]).name
            try:
                Server.request(name)
            except Exception, e:
                self.log.exception(e)
                raise
            # TODO: make more thread-safe
            try:
                self.log.info("Looking up PFN for %s." % path)
                pfn = LookupServer.getPFN(name, protocol='srmv2')
                self.log.info("Looking up TURL for %s." % pfn)
                turl = LookupServer.acquireTURL(pfn)
            except:
                return -errno.EACCES
            self.log.info("Path %s has TURL %s." % (path, turl))
        elif len(parts) > 1 and parts[1] == 'lfns':
            lfn = '/' + '/'.join(parts[2:])
            try: 
                Server.request(lfn)
            except Exception, e:
                self.log.exception(e)
                raise
            try:
                self.log.info("Looking up PFN for %s." % path)
                pfn = LookupServer.getPFN(lfn, protocol='srmv2')
                self.log.info("Looking up TURL for %s." % pfn)
                turl = LookupServer.acquireTURL(pfn)
            except:
                return -errno.EACCES
            self.log.info("Path %s has TURL %s." % (path, turl))

    def read(self, path, length, offset):
        parts = path.split('/')
        assert len(parts[0]) == 0
        assert len(parts) > 2
        if parts[1] == 'runs':
            return self.read_run(parts[2:], length, offset)
        elif parts[1] == 'sites':
            return self.read_sites(parts[2:], length, offset)
        elif parts[1] == 'guids':
            assert len(parts) == 3
            return self.read_guid(parts[2], length, offset)
        elif parts[1] == 'lfns':
            return self.read_lfn('/' + '/'.join(parts[2:]), length, offset)

    def get_dbs_file(self, guid, lfn=False):
        self._dbs_lock.acquire()
        file = None
        try:
            if lfn:
                file = self._dbs.getlfn(guid)
            else:
                file = self._dbs.getguid(guid)
        finally:
            self._dbs_lock.release()
        if not file:
            raise ValueError("Unknown size for file %s." % file)
        return file

    def read_run(self, parts, length, offset):
        """
        Read a single event out of a run, where the event is specified
        by the path.
        """
        assert len(parts) >= 4
        run, event, lumi = parts[-3:]
        dataset = '/'.join([''] + parts)
        res = self._dbs.getFiles(dataset=dataset, run=run, lumi=lumi)
        raise NotImplementedError()

    def read_sites(self, parts, length, offset):
        """
        Read a file where the path is of the form:
        /site_name/store/foo/bar/blah/
        """
        raise NotImplementedError()

    def read_guid(self, guid, length, offset):
        """
        Read a file where the path is given as a GUID.
        """
        return self.read_lfn(guid, length, offset, lfn=False)

    def check_file_manager(self, name, offset, length):
        """
        Check the file manager if it can provide the requested download.
        """
        try:
            pfn = Server.getPfn(name)
        except:
            pfn = None
        if Server.status(name)[0] == 0 or (pfn and os.path.exists(pfn) and \
                os.stat(pfn).st_size < offset+length):
            log.info("HIT: %s in FileManager" % name)
            pfn = Server.getPfn(name)
            self.log.info("LFN already exists in the FileManager.")
            if not os.path.exists(pfn) or os.stat(pfn).st_size < offset+length:
                self.log.info("LFN exists in FileManager, but is invalid at " \
                    "%s." % pfn)
                log.info("INVALID: %s not good in FileManager." % name)
            else:
                try:
                    fp = open(pfn)
                    fp.seek(offset)
                    return fp.read(length)
                except Exception, e:
                    self.log.exception(e)
        elif StatusCode.isFailure(Server.status(name)[0]):
            Server.cancel(lfn, user='root')
            try:
                Server.request(name)
            except Exception, e:
                self.log.exception(e)
                raise
        else:
            self.log.debug("LFN not found in FileManager.")
            log.info("MISS: %s not in FileManager." % name)
        return None

    def read_lfn(self, guid, length, offset, lfn=True):
        dbs_file = self.get_dbs_file(guid, lfn=lfn)
        dbs_file_size = dbs_file.size 
        if length + offset > dbs_file_size:
            length = dbs_file_size - offset
        if offset == dbs_file_size:
            return ''
        name = dbs_file.name
        fm_results = self.check_file_manager(name, offset, length)
        if fm_results:
            return fm_results
        if lfn:
            guid = guid.split('/')[-1].replace('.root', '')
        partial_file = self._pf_factory(guid, name, size=dbs_file_size)
        block_request = BlockRequest(offset, length)
        gridftp = GridftpPartialRunner(block_request, partial_file)
        if not gridftp.can_satisfy:
            self.partial_pool.queue(gridftp)
        results = None
        while results == None:
            results = gridftp.wait_on_data(1)
            if results != None:
                break
            results = self.check_file_manager(name, offset, length)
        if isinstance(results, array.array):
            return results.tostring()
        else:
            return results

    def readlink ( self, path ):
        """
        readlink; not functional
        """
        fuse_log.info('readlink %s' % path)
        return -errno.ENOSYS

    def release ( self, path, flags ):
        """
        Close the file.
        """
        fuse_log.info('release %s %s' % (path, flags))
        try:
            # TODO: fix open/release mechanism
            pass
            #pfn = LookupServer.getPFN(, protocol='srmv2')
            #LookupServer.releaseTURL(pfn)
        except Exception, e:
            self.log.exception(e)
        return -errno.ENOSYS

    def rename ( self, oldPath, newPath ):
        print '*** rename', oldPath, newPath
        return -errno.ENOSYS

    def rmdir ( self, path ):
        print '*** rmdir', path
        return -errno.ENOSYS

    def statfs ( self ):
        print '*** statfs'
        return -errno.ENOSYS

    def symlink ( self, targetPath, linkPath ):
        print '*** symlink', targetPath, linkPath
        return -errno.ENOSYS

    def truncate ( self, path, size ):
        """
        truncate; not functional
        """
        fuse_log.info("truncate %s %s" % (path, size))
        return -errno.ENOSYS

    def unlink ( self, path ):
        """
        unlink; not functional
        """
        fuse_log.info("unlink %s" % path)
        return -errno.ENOSYS

    def utime ( self, path, times ):
        """
        Return utime; not functional
        """
        fuse_log.info('utime %s %s' % (path, times))
        return -errno.ENOSYS

    def write ( self, path, buf, offset ):
        """
        Write something to CmsFs.  Currently a no-op
        """
        fuse_log.info('write %s %s %s' % (path, buf, offset))
        return -errno.ENOSYS

def makeConfigParser():
    cp = ConfigParser.ConfigParser()
    return cp

def main():
    cp = makeConfigParser()
    fs = CmsFsFUSE(cp)
    fs.configure(cp)
    fs.flags = 0
    fs.multithreaded = 1
    fs.parse()
    fs.main()

if __name__ == '__main__':
    main()

