
"""
The GridFTP-block portion of the CmsFs filesystem.
"""

import os
import logging
import threading

from pyGlobus.ftpClient import FtpClient, HandleAttr #pylint: disable-msg=F0401
from pyGlobus.ftpClient import OperationAttr #pylint: disable-msg=F0401
from pyGlobus.ftpClient import RestartMarker #pylint: disable-msg=F0401
from pyGlobus.util import Buffer, GlobusException #pylint: disable-msg=F0401

from ConfiguredObject import ConfiguredObject

from CmsFsTransfers import Block
from FileLookup import LookupServer

log = logging.getLogger("CmsFs.Cache")

class GridftpPartialRunner(ConfiguredObject):

    """
    This class is responsible for filling up partial blocks based upon the
    request and the current contents of the partial_file.

    It is responsible for resolving the LFN to a SURL, then the SURL to a TURL.
    Once this is done, it performs the partial GridFTP transfer
    """

    def __init__(self, block_request, partial_file):
        super(GridftpPartialRunner, self).__init__()
        self.block = partial_file.calculate_partial_block(block_request)
        block_store = partial_file.block_store
        self.block_request = block_request
        self._condExternal = threading.Condition(threading.Lock())
        if isinstance(self.block, Block): # We have a complete block on disk.
            # We don't need to serialize this access, as only one thread can be
            # in the constructor at a time.
            if not self.block.data_loaded():
                block_store.retrieve(self.block, partial_file)
                self.log.debug("GridftpRunner can be fulfilled by an existing" \
                    " block which we just retrieved.")
            else:
                self.log.debug("GridftpRunner can be fulfilled by an existing "\
                    "block.")
            self.can_satisfy = True
            log.info("HIT: %s, %s" % (partial_file, block_request))
            return
        self.log.debug("Starting GridftpRunner for %s." % self.block_request)
        if block_request.partial_block_can_satisfy(self.block):
            self.log.debug("Trivial!  Can already be satisfied.")
            self.can_satisfy = True
            log.info("HIT: %s, %s" % (partial_file, block_request))
            return
        log.info("MISS: %s, %s" % (partial_file, block_request))
        self.can_satisfy = False
        self.cb_handle = None
        self._buffer = Buffer(256*1024)
        self.handleAttr = HandleAttr()
        self._ftpClient = FtpClient(self.handleAttr)
        self._condV = threading.Condition(threading.Lock())
        self.cp = partial_file.cp
        self.partial_file = partial_file
        self.error = None
        self.handle = None
        self.src_url = None

    def wait_on_data(self, timeout=None):
        """
        This function blocks until the block_request is fulfilled or
        there is some GridFTP error.
        """
        if (isinstance(self.block, Block) and self.block_request.\
                block_can_satisfy(self.block)) or self.block_request.\
                partial_block_can_satisfy(self.block):
            return self.block.fulfill_request(self.block_request)
        self._condExternal.acquire()
        self._condExternal.wait(timeout=timeout)
        self._condExternal.release()
        if self.error:
            raise self.error #pylint: disable-msg=W0706
        elif self.block_request.partial_block_can_satisfy(self.block):
            return self.block.fulfill_request(self.block_request)
        elif timeout:
            return None
        else:
            #raise Exception("Improperly signaled completion")
            return self.block.data[self.block_request.offset - \
                self.block.offset:]

    def read_cb(self, arg, handle, buf, bufHandle, bufLen, offset, eof,
            error):
        """
        The read callback function from the GridFTP server.
        """
        if error[0] != 0:
            self.error = Exception("GridFTP error: %s" % str(error))
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()
            raise self.error #pylint: disable-msg=W0706
        self.block.write(buf)
        if self.block_request.partial_block_can_satisfy(self.block):
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()
        if not eof:
            try:
                handle = self._ftpClient.register_read(self._buffer,
                    self.read_cb, None)
                if self.handle != None:
                    self._ftpClient.free_callback(self.handle)
                self.handle = handle
            except Exception, ex:
                self.error = ex
                self._condExternal.acquire()
                self._condExternal.notifyAll()
                self._condExternal.release()
                raise self.error #pylint: disable-msg=W0706
        else:
            # TODO: hack to shorten the block request.
            self.log.debug("Shortening the block request! Requested %s, " \
                "current %s." % (self.block.requested, self.block.current))
            self.block.requested = self.block.current
            self.block.complete = True
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()

    def done_func(self, cv, _, error):
        """
        Cleanup at end of the GridFTP transfer.
        """
        if error[0] != 0:
            self.log.error("The following error occured: %s" % str(error))
            self.error = Exception("GridFTP error: %s" % str(error))
            raise self.error #pylint: disable-msg=W0706
        else:
            self.log.info("No Errors")
            self.block.finish()
        cv.acquire()
        cv.notifyAll()
        cv.release()
        if error[0] != 0:
            raise self.error #pylint: disable-msg=W0706

    def _lookupTURL(self, SURL):
        """
        Return the corresponding gsiftp TURL for a given SURL.
        """
        options = "-T srmv2 -b -p gsiftp"
        cmd = "lcg-getturls %s %s" % (options, SURL)
        fd = os.popen(cmd)
        turl = fd.read()
        if fd.close():
            raise ValueError("Unable to get TURL for SURL %s." % SURL)
        return turl.strip()

    def start(self):
        """
        Start the GridFTP transfer.
        Actually, this is a wrapper around the actual start function in order
        to make sure that any exception is caught and any consumers are
        notified of the error.
        """
        try:
            self._start_real()
        except Exception, error:
            self.error = Exception("GridFTP error: %s" % str(error))
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()
            raise

    def _start_real(self):
        """
        The actual implementation of the start of the GridFTP transfer.
        """
        if (isinstance(self.block, Block) and self.block_request.\
                 block_can_satisfy(self.block)) or self.block_request.\
                 partial_block_can_satisfy(self.block):
            return
        lfn = self.block.parent_file.lfn
        self.log.info("Start Gridftp transfer for LFN %s." % lfn)
        pfn = self.partial_file.getPfn()
        self.log.info("Returned PFN %s for LFN %s." % (pfn, lfn))
        src_url = LookupServer.acquireTURL(pfn)
        self.src_url = src_url
        self.log.info("Returned TURL %s for PFN %s." % (src_url, pfn))
        try:
            operationAttr = OperationAttr()
            restartMarker = RestartMarker()
            self.log.info("Offset %i, length %i" % (self.block.offset,
                self.block.requested))
            self._ftpClient.partial_get(src_url, operationAttr, restartMarker,
                self.block.offset, self.block.offset + self.block.requested,
                self.done_func, self._condV)
            self.handle = self._ftpClient.register_read(self._buffer,
                self.read_cb, None)
        except GlobusException, _:
            if self.src_url:
                LookupServer.releaseTURL(self.src_url)
                self.src_url = None
            raise

        self._condV.acquire()
        self._condV.wait()
        self._condV.release()
        #fd.close()
        if self.src_url:
            LookupServer.releaseTURL(self.src_url)
            self.src_url = None
        self.log.info("Finished Gridftp read for %s cleanly." % src_url)
        if self.handle:
            self._ftpClient.free_callback(self.handle)

    def cancel(self):
        """
        Cancel an in-progress GridFTP transfer
        """
        if self.src_url:
            LookupServer.releaseTURL(self.src_url)
            self.src_url = None
        if not self.block.complete:
            self._ftpClient.abort()


